import logging
from xaal.schemas import devices
from xaal.lib import tools

from .bindings import funct
from functools import partial

logger = logging.getLogger(__name__)


class KNXDev:
    def __init__(self, gw, cfg):
        # print(extract_classname(self))
        self.gateway = gw
        self.cfg = cfg
        self.attributes_binding = {}
        # search for xaal address, if None, the device will produce a new one
        self.addr = tools.get_uuid(cfg.get('addr', None))
        if self.addr == None:
            self.addr = tools.get_random_uuid()
            cfg['addr'] = str(self.addr)
            gw.save_config = True
        self.dev = None
        self.setup()

    def setup(self):
        logger.warn("Please define setup() in this device")

    def write(self, group_addr, dpt, data):
        """ return a function to be call to write the data to right group_addr """
        func = partial(self.gateway.knx.write, group_addr, data, dpt)
        return func

    def bind_attribute(self, attribute, group_addr, func, dpt):
        """ bind a group_addr to a xaal attribute, and apply the func """
        ptr = partial(func, attribute, dpt, data=None)
        self.attributes_binding.update({group_addr: ptr})

    def parse(self, cemi):
        if cemi.group_addr in self.attributes_binding:
            func = self.attributes_binding[cemi.group_addr]
            func(data=cemi.data)

# =============================================================================
# PowerRelay / Lamp ..
# =============================================================================


class OnOffMixin:
    def setup_onoff(self, state_attribute):
        cmd = self.cfg.get('cmd', None)
        if cmd:
            self.dev.add_method('turn_on', self.write(cmd, '1', 1))
            self.dev.add_method('turn_off', self.write(cmd, '1', 0))
        state = self.cfg.get('state', None) or cmd
        mod = self.cfg.get('mod', 'bool')
        if state:
            self.bind_attribute(state_attribute, state, funct[mod], '1')
        self.dev.info = "KNX %s" % cmd or state


class PowerRelay(KNXDev, OnOffMixin):
    def setup(self):
        self.dev = devices.powerrelay(self.addr)
        self.setup_onoff(self.dev.get_attribute("power"))


class PowerRelayToggle(KNXDev, OnOffMixin):
    def setup(self):
        self.dev = devices.powerrelay_toggle(self.addr)
        self.setup_onoff(self.dev.get_attribute("power"))
        toggle = self.cfg.get('toggle', None)
        if toggle:
            self.dev.add_method('toggle', self.write(toggle, '1', 1))


class Lamp(KNXDev, OnOffMixin):
    def setup(self):
        self.dev = devices.lamp(self.addr)
        self.setup_onoff(self.dev.get_attribute("light"))


class LampToggle(KNXDev, OnOffMixin):
    def setup(self):
        self.dev = devices.lamp_toggle(self.addr)
        self.setup_onoff(self.dev.get_attribute("light"))
        toggle = self.cfg.get('toggle', None)
        if toggle:
            self.dev.add_method('toggle', self.write(toggle, '1', 1))


class Switch(KNXDev):
    def setup(self):
        self.dev = devices.switch(self.addr)
        state = self.cfg.get('state', None)
        if state:
            self.bind_attribute(self.dev.get_attribute('position'), state, funct['bool'], '1')
            self.dev.info = "KNX %s" % state


class Shutter(KNXDev):
    def setup(self):
        self.dev = devices.shutter(self.addr)
        up_down = self.cfg.get('updown_cmd', None)
        stop = self.cfg.get('stop_cmd', None)
        if up_down:
            self.dev.add_method('up', self.write(up_down, '1', 0))
            self.dev.add_method('down', self.write(up_down, '1', 1))
        if stop:
            self.dev.add_method('stop', self.write(stop, '1', 1))


class ShutterPosition(KNXDev):
    def setup(self):
        self.dev = devices.shutter_position(self.addr)
        up_down = self.cfg.get('updown_cmd', None)
        stop = self.cfg.get('stop_cmd', None)
        position = self.cfg.get('position_cmd', None)
        if up_down:
            self.dev.add_method('up', self.write(up_down, '1', 0))
            self.dev.add_method('down', self.write(up_down, '1', 1))
        if stop:
            self.dev.add_method('stop', self.write(stop, '1', 1))
        if position:
            self.dev.add_method('position', self.write(position, '1', 1))


# =============================================================================
# Sensors
# =============================================================================
class PowerMeter(KNXDev):
    def setup(self):
        self.dev = devices.powermeter(self.addr)
        self.dev.unsupported_attributes = ['devices']
        self.dev.del_attribute(self.dev.get_attribute('devices'))
        power = self.cfg.get('power', None)
        p_dpt = self.cfg.get('power_dpt', '9')
        p_mod = self.cfg.get('power_mod', 'set')
        if power:
            self.bind_attribute(self.dev.get_attribute('power'), power, funct[p_mod], p_dpt)
        else:
            self.dev.del_attribute(self.dev.get_attribute('power'))
        energy = self.cfg.get('energy', None)
        e_dpt = self.cfg.get('energy_dpt', '13')
        e_mod = self.cfg.get('energy_mod', 'set')
        if energy:
            self.bind_attribute(self.dev.get_attribute('energy'), energy, funct[e_mod], e_dpt)
        else:
            self.dev.del_attribute(self.dev.get_attribute('energy'))
        self.dev.info = "KNX %s" % (power or energy)


class Thermometer(KNXDev):
    def setup(self):
        self.dev = devices.thermometer(self.addr)
        temperature = self.cfg.get('temperature', None)
        t_dpt = self.cfg.get('temperature_dpt', '9')
        t_mod = self.cfg.get('temperature_mod', 'set')
        if temperature:
            self.bind_attribute(self.dev.get_attribute('temperature'), temperature, funct[t_mod], t_dpt)
        self.dev.info = "KNX %s" % temperature


class Hygrometer(KNXDev):
    def setup(self):
        self.dev = devices.hygrometer(self.addr)
        humidity = self.cfg.get('humidity', None)
        h_dpt = self.cfg.get('humidity_dpt', '5.001')
        h_mod = self.cfg.get('humidity_mod', 'round')
        if humidity:
            self.bind_attribute(self.dev.get_attribute('humidity'), humidity, funct[h_mod], h_dpt)
        self.dev.info = "KNX %s" % humidity


class CO2Meter(KNXDev):
    def setup(self):
        self.dev = devices.co2meter(self.addr)
        co2 = self.cfg.get('co2', None)
        c_dpt = self.cfg.get('co2_dpt', '9')
        c_mod = self.cfg.get('co2_mod', 'round')
        if co2:
            self.bind_attribute(self.dev.get_attribute('co2'), co2, funct[c_mod], c_dpt)
        self.dev.info = "KNX %s" % co2


class Luxmeter(KNXDev):
    def setup(self):
        self.dev = devices.luxmeter(self.addr)
        illuminance = self.cfg.get('illuminance', None)
        l_dpt = self.cfg.get('illuminance_dpt', '7')
        l_mod = self.cfg.get('illuminance_mod', 'set')
        if illuminance:
            self.bind_attribute(self.dev.get_attribute('illuminance'), illuminance, funct[l_mod], l_dpt)
        self.dev.info = "KNX %s" % illuminance


class Lightgauge(KNXDev):
    def setup(self):
        self.dev = devices.lightgauge(self.addr)
        brightness = self.cfg.get('brightness', None)
        b_dpt = self.cfg.get('brightness_dpt', '5.001')
        b_mod = self.cfg.get('brightness_mod', 'round')
        if brightness:
            self.bind_attribute(self.dev.get_attribute('brightness'), brightness, funct[b_mod], b_dpt)
        self.dev.info = "KNX %s" % brightness


class Motion(KNXDev):
    def setup(self):
        self.dev = devices.motion(self.addr)
        state = self.cfg.get('state', None)
        if state:
            self.bind_attribute(self.dev.get_attribute('presence'), state, funct['bool'], '1')
        self.dev.info = "KNX %s" % state



#!/usr/bin/env python3
# vim: set encoding=utf-8 tabstop=4 softtabstop=4 shiftwidth=4 expandtab
#########################################################################
#  Copyright 2012-2013 Marcus Popp                         marcus@popp.mx
#########################################################################
#  This file is part of SmartHome.py.
#
#  SmartHome.py is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  SmartHome.py is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with SmartHome.py.  If not, see <http://www.gnu.org/licenses/>.
#########################################################################

import struct
import datetime


def en1(value):
    return [int(value) & 0x01]


def de1(payload):
    if len(payload) != 1:
        return None
    return bool(payload[0] & 0x01)


def en2(payload):
    # control, value
    return [(payload[0] << 1) & 0x02 | payload[1] & 0x01]


def de2(payload):
    if len(payload) != 1:
        return None
    return [payload[0] >> 1 & 0x01, payload[0] & 0x01]


def en3(vlist):
    # direction, value
    return [(int(vlist[0]) << 3) & 0x08 | int(vlist[1]) & 0x07]


def de3(payload):
    if len(payload) != 1:
        return None
    # up/down, stepping
    return [payload[0] >> 3 & 0x01, payload[0] & 0x07]


def en4002(value):
    if isinstance(value, str):
        value = value.encode('iso-8859-1')
    else:
        value = str(value)
    return [0, ord(value) & 0xff]


def de4002(payload):
    if len(payload) != 1:
        return None
    return payload.decode('iso-8859-1')


def en5(value):
    if value < 0:
        value = 0
    elif value > 255:
        value = 255
    return [0, int(value) & 0xff]


def de5(payload):
    if len(payload) != 1:
        return None
    return struct.unpack('>B', payload)[0]


def en5001(value):
    if value < 0:
        value = 0
    elif value > 100:
        value = 100
    return [0, int(value * 255.0 / 100) & 0xff]


def de5001(payload):
    if len(payload) != 1:
        return None
    return struct.unpack('>B', payload)[0] * 100.0 / 255


def en6(value):
    if value < -128:
        value = -128
    elif value > 127:
        value = 127
    return [0, struct.pack('b', int(value))[0]]


def de6(payload):
    if len(payload) != 1:
        return None
    return struct.unpack('b', payload)[0]


def en7(value):
    print(value)
    ret = bytearray()
    ret.extend(struct.pack('>H', int(value)))
    return ret


def de7(payload):
    if len(payload) != 2:
        return None
    return struct.unpack('>H', payload)[0]


def en8(value):
    if value < -32768:
        value = -32768
    elif value > 32767:
        value = 32767
    ret = bytearray()
    ret.extend(struct.pack('>h', int(value)))
    return ret


def de8(payload):
    if len(payload) != 2:
        return None
    return struct.unpack('>h', payload)[0]


def en9(value):
    s = 0
    e = 0
    if value < 0:
        s = 0x8000
    m = int(value * 100)
    while (m > 2047) or (m < -2048):
        e = e + 1
        m = m >> 1
    num = s | (e << 11) | (int(m) & 0x07ff)
    return en7(num)


def de9(payload):
    if len(payload) != 2:
        return None
    i1 = payload[0]
    i2 = payload[1]
    s = (i1 & 0x80) >> 7
    e = (i1 & 0x78) >> 3
    m = (i1 & 0x07) << 8 | i2
    if s == 1:
        s = -1 << 11
    f = (m | s) * 0.01 * pow(2, e)
    return round(f, 2)


def en10(dt):
    return [0, (dt.isoweekday() << 5) | dt.hour, dt.minute, dt.second]


def de10(payload):
    h = payload[0] & 0x1f
    m = payload[1] & 0x3f
    s = payload[2] & 0x3f
    return datetime.time(h, m, s)


def en11(date):
    return [0, date.day, date.month, date.year - 2000]


def de11(payload):
    d = payload[0] & 0x1f
    m = payload[1] & 0x0f
    y = (payload[2] & 0x7f) + 2000  # sorry no 20th century...
    return datetime.date(y, m, d)


def en12(value):
    if value < 0:
        value = 0
    elif value > 4294967295:
        value = 4294967295
    ret = bytearray()
    ret.extend(struct.pack('>I', int(value)))
    return ret


def de12(payload):
    if len(payload) != 4:
        return None
    return struct.unpack('>I', payload)[0]


def en13(value):
    if value < -2147483648:
        value = -2147483648
    elif value > 2147483647:
        value = 2147483647
    ret = bytearray()
    ret.extend(struct.pack('>i', int(value)))
    return ret


def de13(payload):
    if len(payload) != 4:
        return None
    return struct.unpack('>i', payload)[0]


def en14(value):
    ret = bytearray()
    ret.extend(struct.pack('>f', int(value)))
    return ret


def de14(payload):
    if len(payload) != 4:
        return None
    return struct.unpack('>f', payload)[0]


def en16000(value):
    enc = bytearray(1)
    enc.extend(value.encode('ascii')[:14])
    enc.extend([0] * (15 - len(enc)))
    return enc


def en16001(value):
    enc = bytearray(1)
    enc.extend(value.encode('iso-8859-1')[:14])
    enc.extend([0] * (15 - len(enc)))
    return enc


def de16000(payload):
    return payload.rstrip(b'0').decode()


def de16001(payload):
    return payload.rstrip(b'0').decode('iso-8859-1')


def en17(value):
    return [0, int(value) & 0x3f]


def de17(payload):
    if len(payload) != 1:
        return None
    return struct.unpack('>B', payload)[0] & 0x3f


def en20(value):
    return [0, int(value) & 0xff]


def de20(payload):
    if len(payload) != 1:
        return None
    return struct.unpack('>B', payload)[0]


def en24(value):
    enc = bytearray(1)
    enc.extend(value.encode('iso-8859-1'))
    enc.append(0)
    return enc


def de24(payload):
    return payload.rstrip(b'\x00').decode('iso-8859-1')


def en232(value):
    return [0, int(value[0]) & 0xff, int(value[1]) & 0xff, int(value[2]) & 0xff]


def de232(payload):
    if len(payload) != 3:
        return None
    return list(struct.unpack('>BBB', payload))

def enpa(pa):
    pa = pa.split('.')
    return [int(pa[0]) << 4 | int(pa[1]), int(pa[2])]

def depa(string):
    if len(string) != 2:
        return None
    pa = struct.unpack(">H", string)[0]
    return "{0}.{1}.{2}".format((pa >> 12) & 0x0f, (pa >> 8) & 0x0f, (pa) & 0xff)


def enga(ga):
    ga = ga.split('/')
    return [int(ga[0]) << 3 | int(ga[1]), int(ga[2])]


def dega(string):
    if len(string) != 2:
        return None
    ga = struct.unpack(">H", string)[0]
    return "{0}/{1}/{2}".format((ga >> 11) & 0x1f, (ga >> 8) & 0x07, (ga) & 0xff)


decode = {
    '1': de1,
    '2': de2,
    '3': de3,
    '4002': de4002,
    '4.002': de4002,
    '5': de5,
    '5001': de5001,
    '5.001': de5001,
    '6': de6,
    '7': de7,
    '8': de8,
    '9': de9,
    '10': de10,
    '11': de11,
    '12': de12,
    '13': de13,
    '14': de14,
    '16000': de16000,
    '16': de16000,
    '16001': de16001,
    '16.001': de16001,
    '17': de17,
    '20': de20,
    '24': de24,
    '232': de232,
    'pa': depa,
    'ga': dega
}

encode = {
    '1': en1,
    '2': en2,
    '3': en3,
    '4002': en4002,
    '4.002': en4002,
    '5': en5,
    '5001': en5001,
    '5.001': en5001,
    '6': en6,
    '7': en7,
    '8': en8,
    '9': en9,
    '10': en10,
    '11': en11,
    '12': en12,
    '13': en13,
    '14': en14,
    '16000': en16000,
    '16': en16000,
    '16001': en16001,
    '16.001': en16001,
    '17': en17,
    '20': en20,
    '24': en24,
    '232': en232,
    'pa' : enpa,
    'ga': enga
}
# DPT: 19, 28




# xaal use multicast cnx too, so reuse it here. 
# TODO : drop this 

from xaal.lib import network 
from enum import Enum

from . import dpts
import logging
import struct
import time

EMPTY_FRAME = [0x6,0x10,0x5,0x30,0x0,0x0,0x29,0x0,0xbc,0xe0,0x0,0x0,0x0,0x0,0x1,0x0,0x0,]


logger = logging.getLogger(__name__)


def hex_dump(tab):
    for k in tab:
        print('0x%x' % k,end=',')
    print()

def bytes_to_int(bytes):
    result = 0
    for b in bytes:
        result = result * 256 + int(b)
    return result

def int_to_bytes(value, length):
    result = []
    for i in range(0, length):
        result.append(value >> (i * 8) & 0xff)
    result.reverse()
    return result


class KNXCommand(Enum):
    READ = 0x00
    RESP = 0x40
    WRITE = 0x80

class KNXcEMIException(Exception):pass

class KNXcEMI:
    def __init__(self,buf):
        self.parse(buf)
        self.check_sanity()

    def parse(self,buf):
        bytes_ = bytearray(buf)
        #MC if len(bytes_) < 17:
        if len(bytes_) < 16:
            raise KNXcEMIException("Packet to short %d" % len(bytes_))

        self.pkt = bytes_
        # header 
        self.header_lenght = bytes_[0]    # 0x6
        self.version       = bytes_[1]    # 0x10
        self.service       = bytes_[2:4]  # 0x0530 ROUTING_INDICATION
        #self.total_lenght
        # body
        self.msg_code        = bytes_[6]     # 0x29
        self.add_info_lenght = bytes_[7]     # 0x0
        self.ctr_field1      = bytes_[8]     # 0xbc
        self.ctr_field2      = bytes_[9]     # 0xe0
        #self.phy_addr
        #self.group_addr
        self.npdu_lenght     = bytes_[14]    # 0x1
        self.udt             = bytes_[15]    # 0x0 (unsused)
        #self.data   # 0x81 => write 1 / 0x80=> write 0 / 0x0 => read / 0x40 => response

    def check_sanity(self):
        if self.header_lenght != 0x06 : raise KNXcEMIException("Wrong header size 0x%x" % self.header_lenght)
        if self.version     != 0x10: raise KNXcEMIException("Wrong version 0x%x" % self.version)
        if self.service     != bytearray([0x05,0x30]): raise KNXcEMIException("Wrong service 0x%x 0x%x" % self.service[0] % self.service[1])
        if self.msg_code    != 0x29 : raise KNXcEMIException("Wrong msg_code 0x%x" % self.msg_code)
        if self.ctr_field1 not in [0xbc,0xb0] : raise KNXcEMIException("Wrong ctr_field1 0x%x" % self.ctr_field1)
        # JKX: ctr_field2 shoub in [0xe0,0xd0,0xc0] but Mael fix that to: 
        if self.ctr_field2 not in [0xe0,0xd0,0xc0,0xb0,0x40,0x50]: raise KNXcEMIException("Wrong ctr_field2 0x%x" % self.ctr_field2)
        # JKX : how could this arrive ? 
        #if self.npdu_lenght != 0x01 : raise KNXcEMIException("Wrong npdu_lenght 0x%x" % self.npdu_lenght)

    @property
    def total_lenght(self):
        return bytes_to_int(self.pkt[4:6])
    
    @total_lenght.setter
    def total_lenght(self,val):
        self.pkt[4:6] = int_to_bytes(val,2)

    @property
    def phy_addr(self):
        return dpts.decode['pa'](self.pkt[10:12])

    @phy_addr.setter
    def phy_addr(self,val):
        self.pkt[10:12] = dpts.encode['pa'](val)

    @property
    def group_addr(self):
        return dpts.decode['ga'](self.pkt[12:14])

    @group_addr.setter
    def group_addr(self,val):
        self.pkt[12:14] = dpts.encode['ga'](val)

    @property
    def command(self):
        try:
            cmd = KNXCommand(self.pkt[16] & 0xc0) # 0xc0 mask for 0x[0|4|8] in pload
        except ValueError:
            raise KNXcEMIException("wrong command pload : 0x%x" % self.pkt[16])
        return cmd

    @command.setter
    def command(self,cmd):
        self.pkt[16] = cmd.value

    @property
    def data(self):
        if self.total_lenght == 17: # last byte only ie: 0x81 => 1
            return bytearray([self.pkt[16]&0x3f])
        else:
            return self.pkt[17:]

    @data.setter
    def data(self,pload):
        if len(pload) == 1:
            self.pkt[16] = self.pkt[16]|pload[0]
        else:
            self.pkt.extend(pload)

    def __str__(self):
        r = "<KNXcEMI at 0x%x " % (id(self))
        r = r + "pa=%s ga=%s code=0x%x cmd=%s data=%s>" % (self.phy_addr,self.group_addr,self.msg_code,str(self.command),str(self.data))
        return r

class KNXConnector:
    def __init__(self,phy_addr='6.6.6'):
        self.phy_addr = phy_addr
        self.nc = network.NetworkConnector('224.0.23.12',3671,10)
        self.nc.connect()

    def _new_cemi(self,group_addr):
        cemi = KNXcEMI(EMPTY_FRAME)
        cemi.phy_addr = self.phy_addr
        cemi.group_addr = group_addr
        return cemi

    def write(self,group_addr,data,dtype='1'):
        """ send a write request to KNX group_addr"""
        cemi = self._new_cemi(group_addr)
        cemi.command = KNXCommand.WRITE
        cemi.data = dpts.encode[dtype](data)
        self.send(cemi)

    def read(self,group_add):
        """ send a read request to KNX group_addr"""
        frame = self._new_cemi(group_add)
        frame.command = KNXCommand.READ
        self.send(frame)

    def send(self,frame):
        frame.total_lenght = len(frame.pkt) # fix the lenght . 
        logger.debug("Sending: " + str(frame))
        #hex_dump(frame.pkt)
        self.nc.send(frame.pkt)

    def receive(self):
        buf = self.nc.receive()
        if buf:
            frame = KNXcEMI(buf)
            if frame.phy_addr != self.phy_addr:
                return frame
        return None


def test():
    c = KNXConnector()
    for i in range(1,5):
        #c.write('0/0/%s' % i,1)
        c.read('0/0/%s' % i)
        time.sleep(1)
    while 1:
        frame = c.receive()
        if frame:
            logger.debug("Receive: " + (str(frame)))

if __name__ == '__main__':
    test()
    


        
        



from gevent import monkey;monkey.patch_all(thread=False)
import gevent

from xaal.lib import tools,Device,Engine
from .knxrouter import KNXConnector,KNXcEMIException
from . import devices

import inspect
import atexit
import logging

PACKAGE_NAME = 'xaal.knx'
logger = logging.getLogger(PACKAGE_NAME)

def device_class_finder():
    """ look trought the devices module to find binding class"""
    BLACK_LIST = ['KNXDev','partial']
    result = {}
    dict_ = devices.__dict__
    for k in dir(devices):
        target = dict_[k]
        if (inspect.isclass(target)) and (k not in BLACK_LIST):
            if str(k).endswith('Mixin'):continue
            result.update({k.upper():target})
    return result

class GW(gevent.Greenlet):
    def __init__(self,engine):
        gevent.Greenlet.__init__(self)
        self.engine = engine
        self.devices = []
        self.save_config = False
        atexit.register(self._exit)
        self.config()
        self.setup()
        self.refresh()

    def config(self):
        cfg = tools.load_cfg(PACKAGE_NAME)
        if not cfg:
            cfg= tools.new_cfg(PACKAGE_NAME)
            cfg['devices'] = []
            logger.warn("Created an empty config file, please add your devices")
            cfg.write()
        self.cfg = cfg

    def setup(self):
        addr = self.cfg['config'].get('phy_addr',None)
        if addr:
            self.knx = KNXConnector(addr)
        else:
            self.knx = KNXConnector()

        devs = self.cfg['devices']
        devices_class = device_class_finder()
        for k in devs:
            type_ = devs[k].get('type',None)
            klass = None
            if type_:            
                klass = devices_class.get(devs[k]['type'].upper(),None)
            if klass:
                dev = klass(self,devs[k])
                xaal_dev = dev.dev
                xaal_dev.vendor_id = 'IHSEV'
                xaal_dev.version = 0.1
                xaal_dev.product_id = 'KNX ' + type_
                self.devices.append(dev)
            else:
                logger.warn("Unkown device type %s %s" % (type_,k))

        l = [k.dev for k in self.devices]
        self.engine.add_devices(l)
        # schedule a refresh of KNX network. usefull for wifi network
        refresh_rate = self.cfg['config'].get('refresh_rate',None)
        if refresh_rate:
            self.engine.add_timer(self.refresh,int(refresh_rate))
        

    def _run(self):
        while 1:
            try:
                cemi = self.knx.receive()
            except KNXcEMIException as e:
                logger.warn("Wrong cemi frame: " +str(e))
                cemi = None
            if cemi:
                logger.debug("Receing: %s" % cemi)
                self.handle(cemi)

    def handle(self,cemi):
        for dev in self.devices:
            if cemi.group_addr in dev.attributes_binding :
                dev.parse(cemi)

    def refresh(self):
        """loop throught the device adress to find values"""
        for dev in self.devices:
            for ga in dev.attributes_binding:
                self.knx.read(ga)

    def _exit(self):
        # TODO : drop the save_config flag, reloading cfg is better
        if self.save_config:
            logger.info('Saving configuration file')
            self.cfg.write()

def setup(eng):
    GW.spawn(eng)
    return True



from . import dpts

# This file include the functions you can apply to data to 
# match and attribute


def bool_(attribute,dpt,data):
    val = dpts.decode[dpt](data)
    attribute.value = bool(val)

def bool_inv(attribute,dpt,data):
    val = dpts.decode[dpt](data)
    attribute.value = not bool(val)

def on_off(attribute,dpt,data):
    val = bool(dpts.decode[dpt](data))
    if val:
        attribute.value = 'ON'
    else:
        attribute.value = 'OFF'

def round_(attribute,dpt,data):
    val = dpts.decode[dpt](data)
    if val!=None:
        attribute.value = round(val)

def set_(attribute,dpt,data):
    val = dpts.decode[dpt](data)
    if val!=None:
        attribute.value = val

def mul1000_(attribute,dpt,data):
    val = dpts.decode[dpt](data)
    if val!=None:
        attribute.value = round(val * 1000)
        
funct = {
    "bool"     : bool_,
    "bool_inv" : bool_inv,
    "on_off"   : on_off,
    "round"    : round_,
    "set"      : set_,
    "mul1000"  : mul1000_,
}
